import math
from functools import lru_cache

import click
from ib_insync import util
from ib_insync.contract import ComboLeg, Contract, Option, Stock, TagValue
from ib_insync.order import LimitOrder

from thetagang.util import (
    account_summary_to_dict,
    count_short_option_positions,
    get_highest_price,
    get_strike_limit,
    get_target_delta,
    midpoint_or_market_price,
    portfolio_positions_to_dict,
    position_pnl,
    wait_n_seconds,
)

from .options import option_dte

# Typically the amount of time needed when waiting on data from the IBKR API.
# Sometimes it can take a while to retrieve data, and it's lazy-loaded by the
# API, so getting this number right is largely a matter of guesswork.
api_response_wait_time = 120


class PortfolioManager:
    def __init__(self, config, ib, completion_future):
        self.orders = []
        self.config = config
        self.ib = ib
        self.completion_future = completion_future
        self.ib.orderStatusEvent += self.orderStatusEvent

    def orderStatusEvent(self, trade):
        if "Filled" in trade.orderStatus.status:
            click.secho(
                f"Order filled, symbol={trade.contract.symbol}",
                fg="green",
            )
        if "Cancelled" in trade.orderStatus.status:
            click.secho(
                f"Order cancelled, symbol={trade.contract.symbol} log={trade.orderStatus.log}",
                fg="red",
            )
        else:
            click.secho(
                f"Order updated, symbol={trade.contract.symbol} status={trade.orderStatus.status}",
                fg="blue",
            )

    def get_calls(self, portfolio_positions):
        return self.get_options(portfolio_positions, "C")

    def get_puts(self, portfolio_positions):
        return self.get_options(portfolio_positions, "P")

    def get_options(self, portfolio_positions, right):
        r = []
        for symbol in portfolio_positions:
            r = r + list(
                filter(
                    lambda p: (
                        isinstance(p.contract, Option)
                        and p.contract.right.startswith(right)
                    ),
                    portfolio_positions[symbol],
                )
            )

        return r

    def wait_for_midpoint_price(self, ticker):
        try:
            wait_n_seconds(
                lambda: util.isNan(ticker.midpoint()),
                lambda: self.ib.waitOnUpdate(timeout=15),
                api_response_wait_time,
            )
        except RuntimeError:
            return False
        return True

    def wait_for_market_price(self, ticker):
        try:
            wait_n_seconds(
                lambda: util.isNan(ticker.marketPrice()),
                lambda: self.ib.waitOnUpdate(timeout=15),
                api_response_wait_time,
            )
        except:
            return False
        return True

    @lru_cache(maxsize=32)
    def get_chains_for_stock(self, stock):
        return self.ib.reqSecDefOptParams(stock.symbol, "", stock.secType, stock.conId)

    @lru_cache(maxsize=32)
    def get_ticker_for_stock(self, symbol, primary_exchange, midpoint=False):
        stock = Stock(symbol, "SMART", currency="USD", primaryExchange=primary_exchange)
        self.ib.qualifyContracts(stock)
        return self.get_ticker_for(stock)

    @lru_cache(maxsize=32)
    def get_ticker_for(self, contract, midpoint=False):
        [ticker] = self.ib.reqTickers(contract)

        if midpoint:
            self.wait_for_midpoint_price(ticker)
        else:
            self.wait_for_market_price(ticker)

        return ticker

    @lru_cache(maxsize=32)
    def get_ticker_list_for(self, contracts):
        ticker_list = self.ib.reqTickers(*contracts)

        try:
            wait_n_seconds(
                lambda: any([util.isNan(t.midpoint()) for t in ticker_list]),
                lambda: self.ib.waitOnUpdate(timeout=15),
                30,
            )
        except RuntimeError:
            pass

        return ticker_list

    def put_is_itm(self, contract):
        ticker = self.get_ticker_for_stock(contract.symbol, contract.primaryExchange)

        return contract.strike >= ticker.marketPrice()

    def put_can_be_rolled(self, put):
        # Ignore long positions, we only roll shorts
        if put.position > 0:
            return False

        # Check if this put is ITM, and if it's o.k. to roll
        if not self.config["roll_when"]["puts"]["itm"] and self.put_is_itm(
            put.contract
        ):
            return False

        dte = option_dte(put.contract.lastTradeDateOrContractMonth)
        pnl = position_pnl(put)

        roll_when_dte = self.config["roll_when"]["dte"]
        roll_when_pnl = self.config["roll_when"]["pnl"]
        roll_when_min_pnl = self.config["roll_when"]["min_pnl"]

        if (
            "max_dte" in self.config["roll_when"]
            and dte > self.config["roll_when"]["max_dte"]
        ):
            return False

        if dte <= roll_when_dte:
            if pnl >= roll_when_min_pnl:
                click.secho(
                    f"  {put.contract.localSymbol} can be rolled because DTE of {dte} is <= {self.config['roll_when']['dte']} and P&L of {round(pnl * 100, 1)}% is >= {round(roll_when_min_pnl * 100, 1)}%",
                    fg="blue",
                )
                return True
            else:
                click.secho(
                    f"  {put.contract.localSymbol} can't be rolled because P&L of {round(pnl * 100, 1)}% is < {round(roll_when_min_pnl * 100, 1)}%",
                    fg="red",
                )

        if pnl >= roll_when_pnl:
            click.secho(
                f"  {put.contract.localSymbol} can be rolled because P&L of {round(pnl * 100, 1)}% is >= {round(roll_when_pnl * 100, 1)}",
                fg="blue",
            )
            return True

        return False

    def call_is_itm(self, contract):
        ticker = self.get_ticker_for_stock(contract.symbol, contract.primaryExchange)

        return contract.strike <= ticker.marketPrice()

    def call_can_be_rolled(self, call):
        # Ignore long positions, we only roll shorts
        if call.position > 0:
            return False

        # Check if this call is ITM, and it's o.k. to roll
        if not self.config["roll_when"]["calls"]["itm"] and self.call_is_itm(
            call.contract
        ):
            return False

        dte = option_dte(call.contract.lastTradeDateOrContractMonth)
        pnl = position_pnl(call)

        roll_when_dte = self.config["roll_when"]["dte"]
        roll_when_pnl = self.config["roll_when"]["pnl"]
        roll_when_min_pnl = self.config["roll_when"]["min_pnl"]

        if (
            "max_dte" in self.config["roll_when"]
            and dte > self.config["roll_when"]["max_dte"]
        ):
            return False

        if dte <= roll_when_dte:
            if pnl >= roll_when_min_pnl:
                click.secho(
                    f"  {call.contract.localSymbol} can be rolled because DTE of {dte} is <= {self.config['roll_when']['dte']} and P&L of {round(pnl * 100, 1)}% is >= {round(roll_when_min_pnl * 100, 1)}%",
                    fg="blue",
                )
                return True
            else:
                click.secho(
                    f"  {call.contract.localSymbol} can't be rolled because P&L of {round(pnl * 100, 1)}% is < {round(roll_when_min_pnl * 100, 1)}%",
                    fg="red",
                )

        if pnl >= roll_when_pnl:
            click.secho(
                f"  {call.contract.localSymbol} can be rolled because P&L of {round(pnl * 100, 1)}% is >= {round(roll_when_pnl * 100, 1)}",
                fg="blue",
            )
            return True

        return False

    def get_symbols(self):
        return [s for s in self.config["symbols"].keys()]

    def filter_positions(self, portfolio_positions):
        symbols = self.get_symbols()
        return [
            item
            for item in portfolio_positions
            if item.account == self.config["account"]["number"]
            and item.contract.symbol in symbols
            and item.position != 0
            and item.averageCost != 0
        ]

    def get_portfolio_positions(self):
        portfolio_positions = self.ib.portfolio()
        return portfolio_positions_to_dict(self.filter_positions(portfolio_positions))

    def initialize_account(self):
        self.ib.reqMarketDataType(self.config["account"]["market_data_type"])

        if self.config["account"]["cancel_orders"]:
            # Cancel any existing orders
            open_trades = self.ib.openTrades()
            for trade in open_trades:
                if trade.isActive() and trade.contract.symbol in self.get_symbols():
                    click.secho(f"Canceling order {trade.order}", fg="red")
                    self.ib.cancelOrder(trade.order)

    def summarize_account(self):
        account_summary = self.ib.accountSummary(self.config["account"]["number"])
        click.echo()
        click.secho("Account summary:", fg="green")
        click.echo()
        account_summary = account_summary_to_dict(account_summary)

        if "NetLiquidation" not in account_summary:
            raise RuntimeError(
                f"Account number {self.config['account']['number']} appears invalid (no account data returned)"
            )

        justified_values = {
            "ExcessLiquidity": f"{float(account_summary['ExcessLiquidity'].value):,.0f}",
            "NetLiquidation": f"{float(account_summary['NetLiquidation'].value):,.0f}",
            "FullMaintMarginReq": f"{float(account_summary['FullMaintMarginReq'].value):,.0f}",
            "BuyingPower": f"{float(account_summary['BuyingPower'].value):,.0f}",
            "TotalCashValue": f"{float(account_summary['TotalCashValue'].value):,.0f}",
            "Cushion": f"{float(account_summary['Cushion'].value) * 100:.1f}%",
        }

        padding = max([len(v) for v in justified_values.values()])
        justified_values = {k: v.rjust(padding) for k, v in justified_values.items()}

        click.secho(
            f"  Net liquidation   = {justified_values['NetLiquidation']}",
            fg="cyan",
        )
        click.secho(
            f"  Excess liquidity  = {justified_values['ExcessLiquidity']}",
            fg="cyan",
        )
        click.secho(
            f"  Full maint margin = {justified_values['FullMaintMarginReq']}",
            fg="cyan",
        )
        click.secho(
            f"  Buying power      = {justified_values['BuyingPower']}",
            fg="cyan",
        )
        click.secho(
            f"  Total cash value  = {justified_values['TotalCashValue']}",
            fg="cyan",
        )
        click.secho(
            f"  Cushion           = {justified_values['Cushion']}",
            fg="cyan",
        )

        portfolio_positions = self.get_portfolio_positions()

        click.echo()
        click.secho("Portfolio positions:", fg="green")
        click.echo()

        position_values = {}

        def is_itm(p):
            if p.contract.right.startswith("C") and self.call_is_itm(p.contract):
                return "*"
            elif p.contract.right.startswith("P") and self.put_is_itm(p.contract):
                return "*"
            return " "

        for symbol in portfolio_positions.keys():
            for p in portfolio_positions[symbol]:
                position_values[p.contract.conId] = {
                    "qty": str(int(p.position)),
                    "mktprice": f"{p.marketPrice:,.2f}",
                    "avgprice": f"{p.averageCost:,.2f}",
                    "value": f"{p.marketValue:,.0f}",
                    "cost": f"{(p.averageCost * p.position):,.0f}",
                    "p&l": f"{(position_pnl(p) * 100):.2f}%",
                    "itm?": is_itm(p),
                }
                if isinstance(p.contract, Option):
                    position_values[p.contract.conId][
                        "avgprice"
                    ] = f"{p.averageCost/float(p.contract.multiplier):,.2f}"
                    position_values[p.contract.conId][
                        "strike"
                    ] = f"{float(p.contract.strike):,.2f}"
                    position_values[p.contract.conId]["dte"] = str(
                        option_dte(p.contract.lastTradeDateOrContractMonth)
                    )
                    position_values[p.contract.conId]["exp"] = str(
                        p.contract.lastTradeDateOrContractMonth
                    )
        padding = {
            "qty": len("Qty"),
            "mktprice": len("MktPrice"),
            "avgprice": len("AvgPrice"),
            "value": len("Value"),
            "cost": len("Cost"),
            "p&l": len("P&L"),
            "strike": len("Strike"),
            "dte": len("DTE"),
            "exp": len("Exp"),
            "itm?": len("ITM?"),
        }
        for _id, p in position_values.items():
            for col, value in p.items():
                padding[col] = max(padding[col], len(value))

        # Print column headers
        def print_col(c):
            return c.rjust(padding[c.lower()])

        click.secho(
            f"           {print_col('Qty')}  {print_col('MktPrice')}  {print_col('AvgPrice')}  {print_col('Value')}  {print_col('Cost')}  {print_col('P&L')}  {print_col('Strike')}  {print_col('DTE')}  {print_col('Exp')}  {print_col('ITM?')}",
            fg="green",
        )

        for symbol in portfolio_positions.keys():
            click.secho(f"  {symbol}:", fg="cyan")
            sorted_positions = sorted(
                portfolio_positions[symbol],
                key=lambda p: option_dte(p.contract.lastTradeDateOrContractMonth)
                if isinstance(p.contract, Option)
                else -1,  # Keep stonks on top
            )

            def pad(col, id):
                return position_values[id][col].rjust(padding[col])

            for p in sorted_positions:
                id = p.contract.conId
                qty = pad("qty", id)
                mktPrice = pad("mktprice", id)
                avgPrice = pad("avgprice", id)
                value = pad("value", id)
                cost = pad("cost", id)
                pnl = pad("p&l", id)
                if isinstance(p.contract, Stock):
                    click.secho(
                        f"    Stock  {qty}  {mktPrice}  {avgPrice}  {value}  {cost}  {pnl}",
                        fg="cyan",
                    )
                elif isinstance(p.contract, Option):
                    strike = pad("strike", id)
                    dte = pad("dte", id)
                    exp = pad("exp", id)
                    itm = pad("itm?", id)

                    def p_or_c(p):
                        return "Call" if p.contract.right.startswith("C") else "Put "

                    click.secho(
                        f"    {p_or_c(p)}   {qty}  {mktPrice}  {avgPrice}  {value}  {cost}  {pnl}  {strike}  {dte}  {exp}  {itm}",
                        fg="cyan",
                    )
                else:
                    click.secho(f"    {p.contract}", fg="cyan")

        return (account_summary, portfolio_positions)

    def manage(self):
        try:
            self.initialize_account()
            (account_summary, portfolio_positions) = self.summarize_account()

            click.echo()
            click.secho("Checking positions...", fg="green")

            self.check_puts(account_summary, portfolio_positions)
            self.check_calls(account_summary, portfolio_positions)

            # Look for lots of stock that don't have covered calls
            self.check_for_uncovered_positions(account_summary, portfolio_positions)

            # Refresh positions, in case anything changed from the ordering above
            portfolio_positions = self.get_portfolio_positions()

            # Check if we have enough buying power to write some puts
            self.check_if_can_write_puts(account_summary, portfolio_positions)

            # Wait for pending orders
            wait_n_seconds(
                lambda: any(
                    "Pending" in trade.orderStatus.status for trade in self.orders
                ),
                lambda: self.ib.waitOnUpdate(timeout=15),
                api_response_wait_time,
            )

            click.echo()
            click.secho("ThetaGang is done, shutting down! Cya next time.", fg="yellow")

        except:
            click.secho("An exception was raised, exiting", fg="red")
            click.secho("Check log for details", fg="red")
            raise

        finally:
            # Shut it down
            self.completion_future.set_result(True)

    def check_puts(self, account_summary, portfolio_positions):
        # Check for puts which may be rolled to the next expiration or a better price
        puts = self.get_puts(portfolio_positions)

        # find puts eligible to be rolled
        rollable_puts = list(filter(lambda p: self.put_can_be_rolled(p), puts))

        total_rollable_puts = math.floor(sum([abs(p.position) for p in rollable_puts]))

        click.echo()
        click.secho(f"{total_rollable_puts} puts can be rolled", fg="magenta")

        self.roll_puts(rollable_puts, account_summary)

    def check_calls(self, account_summary, portfolio_positions):
        # Check for calls which may be rolled to the next expiration or a better price
        calls = self.get_calls(portfolio_positions)

        # find calls eligible to be rolled
        rollable_calls = list(filter(lambda p: self.call_can_be_rolled(p), calls))
        total_rollable_calls = math.floor(
            sum([abs(p.position) for p in rollable_calls])
        )

        click.echo()
        click.secho(f"{total_rollable_calls} calls can be rolled", fg="magenta")

        self.roll_calls(rollable_calls, account_summary, portfolio_positions)

    def get_maximum_new_contracts_for(self, symbol, primary_exchange, account_summary):
        total_buying_power = self.get_buying_power(account_summary)
        max_buying_power = (
            self.config["target"]["maximum_new_contracts_percent"] * total_buying_power
        )
        ticker = self.get_ticker_for_stock(
            symbol,
            primary_exchange,
        )

        return max([1, round((max_buying_power / ticker.marketPrice()) // 100)])

    def check_for_uncovered_positions(self, account_summary, portfolio_positions):
        for symbol in portfolio_positions:
            call_count = max(
                [0, count_short_option_positions(symbol, portfolio_positions, "C")]
            )
            stock_count = math.floor(
                sum(
                    [
                        p.position
                        for p in portfolio_positions[symbol]
                        if isinstance(p.contract, Stock)
                    ]
                )
            )
            strike_limit = math.ceil(
                max(
                    [get_strike_limit(self.config, symbol, "C") or 0]
                    + [
                        p.averageCost
                        for p in portfolio_positions[symbol]
                        if isinstance(p.contract, Stock)
                    ]
                )
            )

            target_calls = max([0, stock_count // 100])
            new_contracts_needed = target_calls - call_count

            maximum_new_contracts = self.get_maximum_new_contracts_for(
                symbol,
                self.get_primary_exchange(symbol),
                account_summary,
            )
            calls_to_write = max(
                [0, min([new_contracts_needed, maximum_new_contracts])]
            )

            if calls_to_write > 0:
                click.secho(
                    f"Will write {calls_to_write} calls, {new_contracts_needed} needed for {symbol}, capped at {maximum_new_contracts}, at or above strike ${strike_limit} (target_calls={target_calls}, call_count={call_count})",
                    fg="green",
                )
                self.write_calls(
                    symbol,
                    self.get_primary_exchange(symbol),
                    calls_to_write,
                    strike_limit,
                )

    def write_calls(self, symbol, primary_exchange, quantity, strike_limit):
        sell_ticker = self.find_eligible_contracts(
            symbol, primary_exchange, "C", strike_limit
        )

        if not self.wait_for_midpoint_price(sell_ticker):
            click.secho(
                "Couldn't get midpoint price for contract={sell_ticker}, skipping for now",
                fg="red",
            )
            return

        # Create order
        order = LimitOrder(
            "SELL",
            quantity,
            round(get_highest_price(sell_ticker), 2),
            algoStrategy="Adaptive",
            algoParams=[TagValue("adaptivePriority", "Patient")],
            tif="DAY",
        )

        # Submit order
        try:
            trade = self.ib.placeOrder(sell_ticker.contract, order)
            self.orders.append(trade)
            click.echo()
            click.secho("Order submitted", fg="green")
            click.secho(f"{trade}", fg="green")
        except RuntimeError as e:
            click.echo()
            click.secho(e, fg="red")
            click.secho(
                "Order trade submission seems to have failed, or a response wasn't received in time. Continuing anyway...",
                fg="yellow",
            )

    def write_puts(self, symbol, primary_exchange, quantity, strike_limit):
        try:
            sell_ticker = self.find_eligible_contracts(
                symbol, primary_exchange, "P", strike_limit
            )
        except RuntimeError as e:
            click.echo()
            click.secho(e, fg="red")
            click.secho(
                f"Finding eligible contracts for {symbol} failed. Continuing anyway...",
                fg="yellow",
            )
            return

        if not self.wait_for_midpoint_price(sell_ticker):
            click.secho(
                "Couldn't get midpoint price for contract={sell_ticker}, skipping for now",
                fg="red",
            )
            return

        # Create order
        order = LimitOrder(
            "SELL",
            quantity,
            round(get_highest_price(sell_ticker), 2),
            algoStrategy="Adaptive",
            algoParams=[TagValue("adaptivePriority", "Patient")],
            tif="DAY",
        )

        # Submit order
        trade = self.ib.placeOrder(sell_ticker.contract, order)
        self.orders.append(trade)
        click.echo()
        click.secho("Order submitted", fg="green")
        click.secho(f"{trade}", fg="green")

    def get_primary_exchange(self, symbol):
        return self.config["symbols"][symbol].get("primary_exchange", "")

    def get_buying_power(self, account_summary):
        return math.floor(
            float(account_summary["NetLiquidation"].value)
            * self.config["account"]["margin_usage"]
        )

    def check_if_can_write_puts(self, account_summary, portfolio_positions):
        # Get stock positions
        stock_positions = [
            position
            for symbol in portfolio_positions
            for position in portfolio_positions[symbol]
            if isinstance(position.contract, Stock)
        ]

        total_buying_power = self.get_buying_power(account_summary)

        click.echo()
        click.secho(
            f"Total buying power: ${total_buying_power:,.0f} at {round(self.config['account']['margin_usage'] * 100, 1)}% margin usage",
            fg="green",
        )
        click.echo()

        stock_symbols = dict()
        for stock in stock_positions:
            symbol = stock.contract.symbol
            stock_symbols[symbol] = stock

        targets = dict()
        target_additional_quantity = dict()

        # Determine target quantity of each stock
        for symbol in self.config["symbols"].keys():
            click.secho(f"  {symbol}", fg="green")

            ticker = self.get_ticker_for_stock(
                symbol, self.get_primary_exchange(symbol)
            )

            current_position = math.floor(
                stock_symbols[symbol].position if symbol in stock_symbols else 0
            )
            click.secho(
                f"    Current position quantity: {current_position} shares",
                fg="cyan",
            )

            targets[symbol] = round(
                self.config["symbols"][symbol]["weight"] * total_buying_power, 2
            )
            click.secho(f"    Target value: ${targets[symbol]:,.0f}", fg="cyan")
            target_quantity = math.floor(targets[symbol] / ticker.marketPrice())
            click.secho(f"    Target share quantity: {target_quantity:,d}", fg="cyan")

            # Current number of short puts
            put_count = count_short_option_positions(symbol, portfolio_positions, "P")

            target_additional_quantity[symbol] = math.floor(
                target_quantity - current_position - 100 * put_count
            )

            click.secho(
                f"    Net quantity: {target_additional_quantity[symbol]:,d} shares, {target_additional_quantity[symbol] // 100} contracts",
                fg="cyan",
            )

        click.echo()

        # Figure out how many additional puts are needed, if they're needed
        for symbol in target_additional_quantity.keys():
            additional_quantity = target_additional_quantity[symbol] // 100
            # NOTE: it's possible there are non-standard option contract sizes,
            # like with futures, but we don't bother handling those cases.
            # Please don't use this code with futures.
            if additional_quantity >= 1:
                maximum_new_contracts = self.get_maximum_new_contracts_for(
                    symbol,
                    self.get_primary_exchange(symbol),
                    account_summary,
                )
                puts_to_write = min([additional_quantity, maximum_new_contracts])
                if puts_to_write > 0:
                    strike_limit = get_strike_limit(self.config, symbol, "P")
                    if strike_limit:
                        click.secho(
                            f"Will write {puts_to_write} puts, {additional_quantity} needed for {symbol}, capped at {maximum_new_contracts}, at or below strike ${strike_limit}",
                            fg="cyan",
                        )
                    else:
                        click.secho(
                            f"Will write {puts_to_write} puts, {additional_quantity} needed for {symbol}, capped at {maximum_new_contracts}",
                            fg="cyan",
                        )
                    self.write_puts(
                        symbol,
                        self.get_primary_exchange(symbol),
                        puts_to_write,
                        strike_limit,
                    )

        return

    def roll_puts(self, puts, account_summary):
        return self.roll_positions(puts, "P", account_summary)

    def roll_calls(self, calls, account_summary, portfolio_positions):
        return self.roll_positions(calls, "C", account_summary, portfolio_positions)

    def roll_positions(self, positions, right, account_summary, portfolio_positions={}):
        for position in positions:
            symbol = position.contract.symbol
            strike_limit = get_strike_limit(self.config, symbol, right)
            if right.startswith("C"):
                strike_limit = math.ceil(
                    max(
                        [strike_limit or 0]
                        + [
                            p.averageCost
                            for p in portfolio_positions[symbol]
                            if isinstance(p.contract, Stock)
                        ]
                    )
                )

            sell_ticker = self.find_eligible_contracts(
                symbol,
                self.get_primary_exchange(symbol),
                right,
                strike_limit,
                excluded_expiration=position.contract.lastTradeDateOrContractMonth,
                excluded_strikes=[position.contract.strike]
            )

            quantity = abs(position.position)
            maximum_new_contracts = self.get_maximum_new_contracts_for(
                symbol,
                self.get_primary_exchange(symbol),
                account_summary,
            )
            dte = option_dte(position.contract.lastTradeDateOrContractMonth)
            roll_when_dte = self.config["roll_when"]["dte"]
            if dte > roll_when_dte:
                quantity = min([quantity, maximum_new_contracts])

            position.contract.exchange = "SMART"
            buy_ticker = self.get_ticker_for(position.contract, midpoint=True)

            price = midpoint_or_market_price(buy_ticker) - midpoint_or_market_price(
                sell_ticker
            )

            # Create combo legs
            comboLegs = [
                ComboLeg(
                    conId=position.contract.conId,
                    ratio=1,
                    exchange="SMART",
                    action="BUY",
                ),
                ComboLeg(
                    conId=sell_ticker.contract.conId,
                    ratio=1,
                    exchange="SMART",
                    action="SELL",
                ),
            ]

            # Create contract
            combo = Contract(
                secType="BAG",
                symbol=symbol,
                currency="USD",
                exchange="SMART",
                comboLegs=comboLegs,
            )

            # Create order
            order = LimitOrder(
                "BUY",
                quantity,
                round(price, 2),
                algoStrategy="Adaptive",
                algoParams=[TagValue("adaptivePriority", "Patient")],
                tif="DAY",
            )

            # Submit order
            trade = self.ib.placeOrder(combo, order)
            self.orders.append(trade)
            click.echo()
            click.secho(
                f"Order submitted, current position={abs(position.position)} quantity to roll={quantity}, dte={dte}, price={round(price,2)}, trade={trade}",
                fg="green",
            )

    def find_eligible_contracts(
        self, symbol, primary_exchange, right, strike_limit, excluded_expiration=None, excluded_strikes=[]
    ):
        click.echo()
        click.secho(
            f"Searching option chain for symbol={symbol} right={right}, this can take a while...",
            fg="green",
        )
        click.echo()
        stock = Stock(symbol, "SMART", currency="USD", primaryExchange=primary_exchange)
        self.ib.qualifyContracts(stock)

        ticker = self.get_ticker_for(stock)
        tickerValue = ticker.marketPrice()

        chains = self.get_chains_for_stock(stock)
        chain = next(c for c in chains if c.exchange == "SMART")

        def valid_strike(strike):
            if strike in excluded_strikes:
                return False
            elif right.startswith("P") and strike_limit:
                return strike <= tickerValue and strike <= strike_limit
            elif right.startswith("P"):
                return strike <= tickerValue
            elif right.startswith("C") and strike_limit:
                return strike >= tickerValue and strike >= strike_limit
            elif right.startswith("C"):
                return strike >= tickerValue
            return False

        chain_expirations = self.config["option_chains"]["expirations"]
        min_dte = option_dte(excluded_expiration) if excluded_expiration else 0

        strikes = sorted(strike for strike in chain.strikes if valid_strike(strike))
        expirations = sorted(
            exp
            for exp in chain.expirations
            if option_dte(exp) >= self.config["target"]["dte"]
            and option_dte(exp) >= min_dte
        )[:chain_expirations]
        rights = [right]

        def nearest_strikes(strikes):
            chain_strikes = self.config["option_chains"]["strikes"]
            if right.startswith("P"):
                return strikes[-chain_strikes:]
            if right.startswith("C"):
                return strikes[:chain_strikes]

        contracts = [
            Option(
                symbol,
                expiration,
                strike,
                right,
                "SMART",
                tradingClass=chain.tradingClass,
            )
            for right in rights
            for expiration in expirations
            for strike in nearest_strikes(strikes)
        ]

        contracts = self.ib.qualifyContracts(*contracts)

        tickers = self.get_ticker_list_for(tuple(contracts))

        # Filter out contracts which don't have a midpoint price
        tickers = [t for t in tickers if not util.isNan(t.midpoint())]

        def open_interest_is_valid(ticker):
            def open_interest_is_not_ready():
                if right.startswith("P"):
                    return util.isNan(ticker.putOpenInterest)
                if right.startswith("C"):
                    return util.isNan(ticker.callOpenInterest)

            try:
                wait_n_seconds(
                    open_interest_is_not_ready,
                    lambda: self.ib.waitOnUpdate(timeout=15),
                    api_response_wait_time,
                )
            except RuntimeError:
                click.secho(
                    f"Timeout waiting on market data for {ticker.contract}. Continuing...",
                    fg="yellow",
                )
                return False
            finally:
                self.ib.cancelMktData(ticker.contract)

            # The open interest value is never present when using historical
            # data, so just ignore it when the value is None
            if right.startswith("P"):
                return (
                    ticker.putOpenInterest
                    >= self.config["target"]["minimum_open_interest"]
                )
            if right.startswith("C"):
                return (
                    ticker.callOpenInterest
                    >= self.config["target"]["minimum_open_interest"]
                )

        def delta_is_valid(ticker):
            return (
                ticker.modelGreeks
                and not util.isNan(ticker.modelGreeks.delta)
                and ticker.modelGreeks.delta is not None
                and abs(ticker.modelGreeks.delta)
                <= get_target_delta(self.config, symbol, right)
            )

        def price_is_valid(ticker):
            return not util.isNan(ticker.midpoint()) or not util.isNan(
                ticker.marketPrice()
            )

        # Filter out tickers without prices
        tickers = [ticker for ticker in tickers if price_is_valid(ticker)]
        # Filter by delta and open interest
        tickers = [ticker for ticker in tickers if delta_is_valid(ticker)]
        # Fetch market data
        tickers = [
            self.ib.reqMktData(ticker.contract, genericTickList="101")
            for ticker in tickers
        ]
        # Fetch open interest
        tickers = [ticker for ticker in tickers if open_interest_is_valid(ticker)]
        # Sort by delta first, then expiry date
        tickers = sorted(
            reversed(sorted(tickers, key=lambda t: abs(t.modelGreeks.delta))),
            key=lambda t: option_dte(t.contract.lastTradeDateOrContractMonth),
        )

        if len(tickers) == 0:
            raise RuntimeError(f"No valid contracts found for {symbol}. Aborting.")

        # Return the first result
        return tickers[0]



DEFAULT_CONFIG = {
    "target": {
        "maximum_new_contracts_percent": 0.05,
        "delta": 0.3,
    },
    "roll_when": {
        "min_pnl": 0.0,
        "calls": {"itm": True},
        "puts": {"itm": False},
    },
}



import math
from datetime import datetime

from ib_insync import util
from ib_insync.contract import Option


def to_camel_case(snake_str):
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x.title() for x in components[1:])


def account_summary_to_dict(account_summary):
    d = dict()
    for s in account_summary:
        d[s.tag] = s
    return d


def portfolio_positions_to_dict(portfolio_positions):
    d = dict()
    for p in portfolio_positions:
        symbol = p.contract.symbol
        if symbol not in d:
            d[symbol] = []
        d[symbol].append(p)
    return d


def position_pnl(position):
    return position.unrealizedPNL / abs(position.averageCost * position.position)


def count_short_option_positions(symbol, portfolio_positions, right):
    if symbol in portfolio_positions:
        return math.floor(
            -sum(
                [
                    p.position
                    for p in portfolio_positions[symbol]
                    if isinstance(p.contract, Option)
                    and p.contract.right.startswith(right)
                    and p.position < 0
                ]
            )
        )

    return 0


def count_long_option_positions(symbol, portfolio_positions, right):
    if symbol in portfolio_positions:
        return math.floor(
            sum(
                [
                    p.position
                    for p in portfolio_positions[symbol]
                    if isinstance(p.contract, Option)
                    and p.contract.right.startswith(right)
                    and p.position > 0
                ]
            )
        )

    return 0


def wait_n_seconds(pred, body, seconds_to_wait, started_at=None):
    if not started_at:
        started_at = datetime.now()
    diff = datetime.now() - started_at
    if diff.seconds > seconds_to_wait:
        raise RuntimeError(
            "Exhausted retries waiting on predicate. This shouldn't happen."
        )
    if pred():
        body()
        wait_n_seconds(pred, body, seconds_to_wait, started_at)


def get_highest_price(ticker):
    # Returns the highest of either the option model price, the midpoint, or the
    # market price. The midpoint is usually a bit higher than the IB model's
    # pricing, but we want to avoid leaving money on the table in cases where
    # the spread might be messed up. This may in some cases make it harder for
    # orders to fill in a given day, but I think that's a reasonable tradeoff to
    # avoid leaving money on the table.
    return max([midpoint_or_market_price(ticker), ticker.modelGreeks.optPrice])


def midpoint_or_market_price(ticker):
    # As per the ib_insync docs, marketPrice returns the last price first, but
    # we often prefer the midpoint over the last price. This function pulls the
    # midpoint first, then falls back to marketPrice() if midpoint is nan.
    if util.isNan(ticker.midpoint()):
        if util.isNan(ticker.marketPrice()):
            # Fallback to the model price
            return ticker.modelGreeks.optPrice
        else:
            return ticker.marketPrice()

    return ticker.midpoint()


def get_target_delta(config, symbol, right):
    p_or_c = "calls" if right.startswith("C") else "puts"
    if (
        p_or_c in config["symbols"][symbol]
        and "delta" in config["symbols"][symbol][p_or_c]
    ):
        return config["symbols"][symbol][p_or_c]["delta"]
    if "delta" in config["symbols"][symbol]:
        return config["symbols"][symbol]["delta"]
    if p_or_c in config["target"]:
        return config["target"][p_or_c]["delta"]
    return config["target"]["delta"]


def get_strike_limit(config, symbol, right):
    p_or_c = "calls" if right.startswith("C") else "puts"
    if (
        p_or_c in config["symbols"][symbol]
        and "strike_limit" in config["symbols"][symbol][p_or_c]
    ):
        return config["symbols"][symbol][p_or_c]["strike_limit"]
    return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['thetagang']

package_data = \
{'': ['*']}

install_requires = \
['click-log>=0.3.2,<0.4.0',
 'click>=7.1.2,<8.0.0',
 'colorama>=0.4.4,<0.5.0',
 'ib_insync>=0.9.66,<0.10.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pytimeparse>=1.1.8,<2.0.0',
 'schema>=0.7.3,<0.8.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['thetagang = thetagang.entry:cli',
                     'vscode = vscode:vscode']}

setup_kwargs = {
    'name': 'thetagang',
    'version': '0.1.27',
    'description': 'ThetaGang is an IBKR bot for getting money',
    'long_description': '[![Docker publish](https://github.com/brndnmtthws/thetagang/workflows/Docker%20publish/badge.svg)](https://hub.docker.com/r/brndnmtthws/thetagang) [![Python Publish](https://github.com/brndnmtthws/thetagang/workflows/Python%20Publish/badge.svg)](https://pypi.org/project/thetagang/) [![Docker Pulls](https://img.shields.io/docker/pulls/brndnmtthws/thetagang)](https://hub.docker.com/r/brndnmtthws/thetagang) [![PyPI download month](https://img.shields.io/pypi/dm/thetagang?label=PyPI%20downloads)](https://pypi.python.org/pypi/thetagang/)\n\n[💬 Join the Matrix chat, we can get money together](https://matrix.to/#/#thetagang:frens.io).\n\n# Θ ThetaGang Θ\n\n*Beat the capitalists at their own game with ThetaGang 📈*\n\n![Decay my sweet babies](thetagang.jpg)\n\nThetaGang is an [IBKR](https://www.interactivebrokers.com/) trading bot for\ncollecting premium by selling options using "The Wheel" strategy. The Wheel\nis a strategy that [surfaced on\nReddit](https://www.reddit.com/r/options/comments/a36k4j/the_wheel_aka_triple_income_strategy_explained/),\nbut has been used by many in the past. This bot implements a slightly\nmodified version of The Wheel, with my own personal tweaks.\n\nI\'ve been streaming most of the work on this project [on Twitch, so follow me\nover there](https://www.twitch.tv/letsmakestuff).\n\n## How it works\n\nStart by reading [the Reddit\npost](https://www.reddit.com/r/options/comments/a36k4j/the_wheel_aka_triple_income_strategy_explained/)\nto get some background.\n\nThe strategy, as implemented here, does a few things differently from the one\ndescribed in the post above. For one, it\'s intended to be used to augment a\ntypical index-fund based portfolio with specific asset allocations. For\nexample, you might want to use a 60/40 portfolio with SPY (S&P500 fund) and\nTLT (20 year treasury fund). This strategy reduces risk, but may also limit\ngains from big market swings. By reducing risk, one can increase leverage.\n\nThe main difference between ThetaGang and simply buying and holding index\nfunds is that this script will attempt to harvest volatility by selling\noptions, rather than buying shares directly. This works because implied\nvolatility is typically higher than realized volatility on average. Instead\nof buying shares, you write puts. This has pros and cons, which are outside\nthe scope of this README.\n\nYou could use this tool on individual stocks, but I personally don\'t\nrecommend it because I am not smart enough to understand which stocks to buy.\nThat\'s why I just buy index funds.\n\nThetaGang will try to acquire your desired allocation of each stock or ETF\naccording to the weights you specify in the config. To acquire the positions,\nthe script will write puts when conditions are met (adequate buying power,\nacceptable contracts are available, enough shares needed, etc).\n\nThetaGang will continue to roll any open option positions indefinitely, with\nthe only exception being ITM puts. Once puts are in the money, they will be\nignored until they expire and are exercised (after which you will own the\nunderlying).\n\nIf puts are exercised due to being ITM at expiration, you will own the\nstock, and ThetaGang switches from writing puts to writing calls at a strike\nat least as high as the average cost of the stock held.\n\nPlease note: this strategy is based on the assumption that implied volatility\nis, on average, always higher than realized volatility. In cases where this\nis not true, this strategy will cause you to lose money.\n\nIn the case of deep ITM calls, the bot will prefer to roll the calls to next\nstrike or expiration rather than allowing the underlying to get called away. If you\ndon\'t have adequate buying power available in your account, it\'s possible\nthat the options may get exercised instead of rolling forward and the process\nstarts back at the beginning. Please keep in mind this may have tax\nimplications, but that is outside the scope of this README.\n\nIn normal usage, you would run the script as a cronjob on a daily, weekly, or\nmonthly basis according to your preferences. Running more frequently than\ndaily is not recommended, but the choice is yours.\n\n![Paper account sample output](sample.gif)\n\n## Project status\n\nThis project is, in its current state, considered to be complete. I\'m open\nto contributions, but I am unlikely to accept PRs or feature requests that\ninvolve significant changes to the underlying algorithm.\n\nIf you find something that you think is a bug, or some other issue, please\n[create a new issue](https://github.com/brndnmtthws/thetagang/issues/new).\n\n## "Show me your gains bro" – i.e., what are the returns?\n\nAs discussed elsewhere in this README, you must conduct your own research,\nand I suggest starting with resources such as CBOE\'s BXM and BXDM indices,\nand comparing those to SPX. I\'ve had a lot of people complain because "that\nstrategy isn\'t better than buy and hold BRUH"–let me assure you, that is not\nmy goal with this.\n\nThere are conflicting opinions about whether selling options is good or bad,\nmore or less risky, yadda yadda, but generally the risk profile for covered\ncalls and naked puts is no worse than the worst case for simply holding an\nETF or stock. In fact, I\'d argue that selling a naked put is better than\nbuying SPY with a limit order, because at least if SPY goes to zero you keep\nthe premium from selling the option. The main downside is that returns are\ncapped on the upside. Depending on your goals, this may not matter. If you\'re\nlike me, then you\'d rather have consistent returns and give up a little bit\nof potential upside.\n\nGenerally speaking, the point of selling options is not to exceed the returns\nof the underlying, but rather to reduce risk. Reducing risk is an important\nfeature because it, in turn, allows one to increase risk in other ways\n(i.e., allocate higher percentage to stocks or buy riskier assets).\n\nWhether you use this or not is up to you. I have not one single fuck to give,\nwhether you use it or not. I am not here to convince you to use it, I merely\nwant to share knowledge and perhaps help create a little bit of wealth\nredistribution.\n\n💫\n\n## Requirements\n\nThe bot is based on the [ib_insync](https://github.com/erdewit/ib_insync)\nlibrary, and uses [IBC](https://github.com/IbcAlpha/IBC) for managing the API\ngateway.\n\nTo use the bot, you\'ll need an Interactive Brokers account with a working\ninstallation of IBC. If you want to modify the bot, you\'ll need an\ninstallation of Python 3.8 or newer with the\n[`poetry`](https://python-poetry.org/) package manager.\n\nOne more thing: to run this on a live account, you\'ll require enough capital\nto purchase at least 100 shares of the stocks or ETFs you choose. For\nexample, if SPY is trading at $300/share you\'d need $30,000 available. You\ncan search for lower priced alternatives, but these tend to have low volume\non options which may not be appropriate for this strategy. You should\ngenerally avoid low volume ETFs/stocks. If you don\'t have that kind of\ncapital, you\'ll need to keep renting out your time to the capitalists until\nyou can become a capitalist yourself. That\'s the way the pyramid scheme we\ncall capitalism works.\n\n## Installation\n\n*Before running ThetaGang, you should set up an IBKR paper account to test the\ncode.*\n\n```console\n$ pip install thetagang\n```\n\nIt\'s recommended you familiarize yourself with\n[IBC](https://github.com/IbcAlpha/IBC) so you know how it works. You\'ll need\nto know how to configure the various knows and settings, and make sure things\nlike API ports are configured correctly. If you don\'t want to mess around too\nmuch, consider [running ThetaGang with Docker](#running-with-docker).\n\n## Usage\n\n```console\n$ thetagang -h\n```\n\n## Up and running with Docker\n\nMy preferred way for running ThetaGang is to use a cronjob to execute Docker\ncommands. I\'ve built a Docker image as part of this project, which you can\nuse with your installation. There\'s a [prebuilt Docker image\nhere](https://hub.docker.com/repository/docker/brndnmtthws/thetagang).\n\nTo run ThetaGang within Docker, you\'ll need to pass `config.ini` for [IBC\nconfiguration](https://github.com/IbcAlpha/IBC/blob/master/userguide.md) and\n[`thetagang.toml`](/thetagang.toml) for ThetaGang. There\'s a sample\n[`ibc-config.ini`](/ibc-config.ini) included in this repo for your convenience.\n\nThe easiest way to get the config files into the container is by mounting a\nvolume.\n\nTo get started, grab a copy of `thetagang.toml` and `config.ini`:\n\n```console\n$ mkdir ~/thetagang\n$ cd ~/thetagang\n$ curl -Lq https://raw.githubusercontent.com/brndnmtthws/thetagang/main/thetagang.toml -o ~/thetagang/thetagang.toml\n$ curl -Lq https://raw.githubusercontent.com/brndnmtthws/thetagang/main/ibc-config.ini -o ~/thetagang/config.ini\n```\n\nEdit `~/thetagang/thetagang.toml` to suit your needs. Pay particular\nattention to the symbols and weights. At a minimum, you must change the\nusername, password, and account number. You may also want to change the\ntrading move from paper to live when needed.\n\nNow, to run ThetaGang with Docker:\n\n```console\n$ docker run --rm -i --net host \\\n    -v ~/thetagang:/etc/thetagang \\\n    brndnmtthws/thetagang:latest \\\n    --config /etc/thetagang/thetagang.toml\n```\n\nLastly, to run ThetaGang as a daily cronjob Monday to Friday at 9am, add\nsomething like this to your crontab (on systems with a cron installation, use\n`crontab -e` to edit your crontab):\n\n```crontab\n0 9 * * 1-5 docker run --rm -i -v ~/ibc:/etc/thetagang brndnmtthws/thetagang:latest --config /etc/thetagang/thetagang.toml\n```\n\n## Determining which ETFs or stocks to run ThetaGang with\n\nI leave this as an exercise to the reader, however I will provide a few\nrecommendations and resources:\n\n### Recommendations\n\n* Stick with high volume ETFs or stocks\n* Careful with margin usage, you\'ll want to calculate the worst case scenario\n  and provide plenty of cushion for yourself based on your portfolio\n\n\n### Resources\n\n* For discussions about selling options, check out\n  [r/thetagang](https://www.reddit.com/r/thetagang/)\n* For backtesting portfolios, you can use [this\n  tool](https://www.portfoliovisualizer.com/backtest-portfolio) and [this\n  tool](https://www.portfoliovisualizer.com/optimize-portfolio) to get an idea\n  of drawdown and typical volatility\n\n## Development\n\nCheck out the code to your local machine and install the Python dependencies:\n\n```shell\n$ poetry install\n$ poetry run autohooks activate\n$ poetry run thetagang -h\n...\n```\n\nYou are now ready to make a splash! 🐳\n\n## FAQ\n\n| Error                                                                                                      | Cause                                                                                                                            | Resolution                                                                                                                                                                                                                                                                                                                                                                                                                                         |\n| ---------------------------------------------------------------------------------------------------------- | -------------------------------------------------------------------------------------------------------------------------------- | -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |\n| Requested market data is not subscribed.                                                                   | Requisite market data subscriptions have not been set up on IBKR.                                                                | [Configure](https://www.interactivebrokers.com/en/software/am3/am/settings/marketdatasubscriptions.htm) your market data subscriptions. The default config that ships with this script uses the `Cboe One Add-On Bundle` and the `US Equity and Options Add-On Streaming Bundle`. **Note**: You _must_ fund your account before IBKR will send data for subscriptions. Without funding you can still subscribe but you will get an error from ibc. |\n| No market data during competing live session                                                               | Your account is logged in somewhere else, such as the IBKR web portal, the desktop app, or even another instance of this script. | Log out of all sessions and then re-run the script.                                                                                                                                                                                                                                                                                                                                                                                                |\n| `ib_insync.wrapper ERROR Error 200, reqId 10: The contract description specified for SYMBOL is ambiguous.` | IBKR needs to know which exchange is the primary exchange for a given symbol.                                                    | You need to specify the primary exchange for the stock. This is normal for companies, typically. For ETFs it usually isn\'t required. Specify the `primary_exchange` parameter for the symbol, i.e., `primary_exchange = "NYSE"`.                                                                                                                                                                                                                   |\n\n## Support and sponsorship\n\nIf you get some value out of this, please consider [sponsoring me](https://github.com/sponsors/brndnmtthws)\nto continue maintaining this project well into the future. Like\neveryone else in the world, I\'m just trying to survive.\n\nIf you like what you see but want something different, I am willing\nto work on bespoke or custom trading bots for a fee. Reach out\nto me directly through my GitHub profile.\n\n## Stargazers over time\n\n[![Stargazers over time](https://starchart.cc/brndnmtthws/thetagang.svg)](https://starchart.cc/brndnmtthws/thetagang)\n',
    'author': 'Brenden Matthews',
    'author_email': 'brenden@brndn.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/brndnmtthws/thetagang',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import math

import click
from schema import And, Optional, Or, Schema

import thetagang.config_defaults as config_defaults
from thetagang.dict_merge import dict_merge


def normalize_config(config):
    # Do any pre-processing necessary to the config here, such as handling
    # defaults, deprecated values, config changes, etc.

    if "twsVersion" in config["ibc"]:
        click.secho(
            "WARNING: config param ibc.twsVersion is deprecated, please remove it from your config.",
            fg="yellow",
            err=True,
        )

        # TWS version is pinned to latest stable, delete any existing config if it's present
        del config["ibc"]["twsVersion"]

    if "maximum_new_contracts" in config["target"]:
        click.secho(
            "WARNING: config param target.maximum_new_contracts is deprecated, please remove it from your config.",
            fg="yellow",
            err=True,
        )

        del config["target"]["maximum_new_contracts"]

    # xor: should have weight OR parts, but not both
    if any(["weight" in s for s in config["symbols"].values()]) == any(
        ["parts" in s for s in config["symbols"].values()]
    ):
        raise RuntimeError(
            "ERROR: all symbols should have either a weight or parts specified, but parts and weights cannot be mixed."
        )

    if "parts" in list(config["symbols"].values())[0]:
        # If using "parts" instead of "weight", convert parts into weights
        total_parts = float(sum([s["parts"] for s in config["symbols"].values()]))
        for k in config["symbols"].keys():
            config["symbols"][k]["weight"] = config["symbols"][k]["parts"] / total_parts
        for s in config["symbols"].values():
            del s["parts"]

    return apply_default_values(config)


def apply_default_values(config):
    return dict_merge(config_defaults.DEFAULT_CONFIG, config)


def validate_config(config):
    if "minimum_cushion" in config["account"]:
        raise "Config error: minimum_cushion is deprecated and replaced with margin_usage. See sample config for details."

    schema = Schema(
        {
            "account": {
                "number": And(str, len),
                "cancel_orders": bool,
                "margin_usage": And(float, lambda n: 0 <= n),
                "market_data_type": And(int, lambda n: 1 <= n <= 4),
            },
            "option_chains": {
                "expirations": And(int, lambda n: 1 <= n),
                "strikes": And(int, lambda n: 1 <= n),
            },
            "roll_when": {
                "pnl": And(float, lambda n: 0 <= n <= 1),
                "dte": And(int, lambda n: 0 <= n),
                "min_pnl": float,
                Optional("max_dte"): And(int, lambda n: 1 <= n),
                Optional("calls"): {
                    "itm": bool,
                },
                Optional("puts"): {
                    "itm": bool,
                },
            },
            "target": {
                "dte": And(int, lambda n: 0 <= n),
                "delta": And(float, lambda n: 0 <= n <= 1),
                Optional("maximum_new_contracts"): And(int, lambda n: 1 <= n),
                Optional("maximum_new_contracts_percent"): And(
                    float, lambda n: 0 <= n <= 1
                ),
                "minimum_open_interest": And(int, lambda n: 0 <= n),
                Optional("calls"): {
                    Optional("delta"): And(float, lambda n: 0 <= n <= 1),
                },
                Optional("puts"): {
                    Optional("delta"): And(float, lambda n: 0 <= n <= 1),
                },
            },
            "symbols": {
                object: {
                    Or("weight", "parts", only_one=True): And(
                        Or(float, int),
                        lambda n: 0 <= n <= 1 if isinstance(n, float) else n > 0,
                    ),
                    Optional("primary_exchange"): And(str, len),
                    Optional("delta"): And(float, lambda n: 0 <= n <= 1),
                    Optional("calls"): {
                        Optional("delta"): And(float, lambda n: 0 <= n <= 1),
                        Optional("strike_limit"): And(float, lambda n: n > 0),
                    },
                    Optional("puts"): {
                        Optional("delta"): And(float, lambda n: 0 <= n <= 1),
                        Optional("strike_limit"): And(float, lambda n: n > 0),
                    },
                }
            },
            Optional("ib_insync"): {Optional("logfile"): And(str, len)},
            "ibc": {
                Optional("password"): And(str, len),
                Optional("userid"): And(str, len),
                Optional("gateway"): bool,
                Optional("RaiseRequestErrors"): bool,
                Optional("ibcPath"): And(str, len),
                Optional("tradingMode"): And(
                    str, len, lambda s: s in ("live", "paper")
                ),
                Optional("ibcIni"): And(str, len),
                Optional("twsPath"): And(str, len),
                Optional("twsSettingsPath"): And(str, len),
                Optional("javaPath"): And(str, len),
                Optional("fixuserid"): And(str, len),
                Optional("fixpassword"): And(str, len),
            },
            "watchdog": {
                Optional("appStartupTime"): int,
                Optional("appTimeout"): int,
                Optional("clientId"): int,
                Optional("connectTimeout"): int,
                Optional("host"): And(str, len),
                Optional("port"): int,
                Optional("probeTimeout"): int,
                Optional("readonly"): bool,
                Optional("retryDelay"): int,
                Optional("probeContract"): {
                    Optional("currency"): And(str, len),
                    Optional("exchange"): And(str, len),
                    Optional("secType"): And(str, len),
                    Optional("symbol"): And(str, len),
                },
            },
        }
    )
    schema.validate(config)

    assert len(config["symbols"]) > 0

    assert math.isclose(
        1, sum([s["weight"] for s in config["symbols"].values()]), rel_tol=1e-5
    )



# Source: https://gist.github.com/angstwad/bf22d1822c38a92ec0a9

# Recursive dictionary merge
# Copyright (C) 2016 Paul Durivage <pauldurivage+github@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import collections


def dict_merge(dct, merge_dct):
    """Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k in merge_dct.keys():
        if (
            k in dct
            and isinstance(dct[k], dict)
            and isinstance(merge_dct[k], collections.Mapping)
        ):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]

    return dct
